// Encoding: ISO-8859-2
// prvocisla.c
// een IJC-DU1, pklad b), 22.3.2010
// Autor: Milan Falenk, FIT, xfales01@stud.fit.vutbr.cz
// Peloeno: gcc 4.4.5 x86-64, Xubuntu linux
// Rychlost programu: 	INLINE/MACRO:
//			~1.5s@powersave(Core i5@1.2GHz)
//			~1.1s@performace(Core i5@2.27GHz)
//			~1.2s@merlin

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include "error.h"

const char errstr[] = "Index %ld mimo rozsah 0..%ld";

// Pocet bitu ulintu
#define ULINT_BITS (sizeof(unsigned long) * 8)
// Vytvori bitovou masku
#define BIT_MASK(x) ((unsigned long)(0x01<<((x)%ULINT_BITS)))
// Velikost pole v jednotlivych polozkach
#define FIELD_SIZE(bits) (((bits)%ULINT_BITS != 0) ? ((bits)/ULINT_BITS) + 1 : (bits)/ULINT_BITS )
#define BitArray(field, size) unsigned long field[FIELD_SIZE(size)]; \
memset(field, 0, FIELD_SIZE(size)*sizeof(unsigned long));\
const unsigned long size_##field = (size)-1;

#ifndef USE_INLINE

//#define GetBit(field, index) ((index > size_##field) ? Error(errstr, (unsigned long) index, (unsigned long) size_##field), 0 : (((field[(index)/ULINT_BITS] & BIT_MASK(index)) == 0) ? 0 : 1 ) )
/*
#define SetBit(field, index, expr) if(index > size_##field) \
Error(errstr, (unsigned long) index, (unsigned long) size_##field); \
if((expr) == 0) \
field[(index)/ULINT_BITS] &= ~BIT_MASK(index); \
else \
field[(index)/ULINT_BITS] |= BIT_MASK(index);
*/

#define GetBit(field, index) (((index) > size_##field) ? Error(errstr, (unsigned long)(index), (unsigned long)size_##field), 0 : ((field[(index)/ULINT_BITS] & BIT_MASK(index)) == 0 ? 0 : 1))
#define SetBit(field, index, expr) { \
if ((index) > size_##field) Error(errstr, (unsigned long)(index), (unsigned long)size_##field); \
if ((expr)==1) field[(index)/ULINT_BITS] |= BIT_MASK(index); \
else field[(index)/ULINT_BITS] &= ~BIT_MASK(index); \
}


#else

#define GetBit(field, index) GetBitInline(field, index, size_##field)
static inline unsigned long GetBitInline(unsigned long *field, unsigned long index, unsigned long  size_field)
{
    if(index > size_field)
	Error(errstr, (unsigned long) index, (unsigned long) size_field);
    
    return ((field[index/ULINT_BITS]&BIT_MASK(index)) == 0 ? 0 : 1 );
}

#define SetBit(field, index, expr) SetBitInline(field, index, (expr), size_##field)
static inline void SetBitInline(unsigned long *field, unsigned long index, int bit, unsigned long size_field)
{
    if(index>size_field)
	Error(errstr, (unsigned long) index, (unsigned long) size_field);
    
    if((bit)==0)
	field[index/ULINT_BITS] &= ~BIT_MASK(index);
    else
	field[index/ULINT_BITS] |= BIT_MASK(index);
}
#endif

int main(void)
{
    // Erastothenovo sto
    const unsigned long N = 90000000L;
    const unsigned long iB = 2; // potek i
    const unsigned long numWrite = 20;
    BitArray(p, N);
    unsigned long sN = round(sqrt((double)N));
    unsigned long mul;
    
    for(unsigned long i = iB; i < sN;) // dokud nejsme u odmocniny
    {
	for(; GetBit(p, i) != 0; i++); // hledame nejmensi i s nulou
	
	for(unsigned long n = 2;  (mul = n*i) < N; n++) // nastavime nasobky na 1
	    SetBit(p, mul, 1);
	
	i++;
    }
    
    // vypis nalezenych cisel
    unsigned long numbers[numWrite];
    int count = 0;
    for(unsigned long i = N-1; count < 20; i--)		// Ulozime si pozadovana cisla do pole
    {
	if(GetBit(p, i) == 0)
	    numbers[numWrite - count++ - 1] = i;
    }
    
    for(unsigned long i = 0; i < numWrite ; i++)	// A vypiseme je
	printf("%lu\n", numbers[i]);

    return EXIT_SUCCESS;
}

